/**************************************************************/
/************* QP.c ****************************************/
/**************************************************************/
/*    Irene Rodriguez-Lujan    *********************************/
/*    Ramon Huerta			  *********************************/
/*    Charles Elkan			  *********************************/
/*    Carlos Santa Cruz		  *********************************/
/**************************************************************/
/** June 3rd, 2010                                        ****/
/** Based on "Quadratic Programming Feature Selection"       **/
/**************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "QP.h"
#include "NYSTROM.h"


/* QP-solver */
void qpgen2_(double *G, double *g0, int *n1, int *n2,double *x,double *crval,double *CI,double *ci0,int *n,int *p,int *m,double *iact,
			 int *nact, int *iter, double *work,int *ierr); 

/* Data discretization */
short QP_DataDiscretization(double **data, long n, long d, long nClasses, long nSegmentos,long nSigma);

/* Numerical rutines */
short QP_CalcEigen(QP_pQP x);
int split_cake2(int d,int n,double *C,double *F,double * *res,float *fr, double min_alo, double max_alo, double **evec);
void QP_WeightedCorrelationRandom(double **C,double *media, double *varianza,double **data,long nData,long d,long nClases,long *aleat);





/***********************************************/
/* QP structure allocation		              */
/***********************************************/


QP_pQP QP_AllocQP(long d,double alpha)
{
  
  QP_pQP	x=NULL;
  long i;

  /* allocation */
  if (!(x=(QP_pQP)calloc(1,sizeof(QP_qp)))) {
    fprintf(stderr,"ERROR: allocation QP structure\n");
    return NULL;
  }

  
  x->d=d;
  x->alpha=alpha;
    
  /* Quadratic term matrix */
  x->Q=(double **) malloc (d*sizeof(double*));

  if (!x->Q)
  {
    QP_FreeQP(&x);
    fprintf(stderr,"ERROR: allocation QP structure (quadratic term matrix %ld x %ld)\n",d,d);
    return NULL;
  }

  for (i=0;i<d;i++)
  {
	x->Q[i]=(double *) malloc (d*sizeof(double));
	if (!x->Q[i])
	{
    		QP_FreeQP(&x);
			fprintf(stderr,"ERROR: allocation QP structure (quadratic term matrix %ld x %ld)\n",d,d);
    		return NULL;
	}

  }
	
  /* Linear term vector */
  x->F=(double *) malloc (d*sizeof(double));
  if (!x->F)
  {
	QP_FreeQP(&x);
	fprintf(stderr,"ERROR: allocation QP structure (linear term vector %ld)\n",d);
    return NULL;
  }


  /* Solution vector */
  x->h=(double *) malloc (d*sizeof(double));
  if (!x->F)
  {
	QP_FreeQP(&x);
	fprintf(stderr,"ERROR: allocation QP structure (solution vector %ld)\n",d);
    return NULL;
  }


  /* eigenvectors of quadratic term */
  x->evec=(double **) malloc (d*sizeof(double *));
  if (!x->evec)
  {
    QP_FreeQP(&x);
    fprintf(stderr,"ERROR: allocation QP structure (eigenvectors matrix %ld x %ld)\n",d,d);
    return NULL;
  }

  for (i=0;i<d;i++)
  {
	x->evec[i]=(double *)malloc(d*sizeof(double));
	if (!x->evec[i])
	{
	  QP_FreeQP(&x);
	  fprintf(stderr,"ERROR: allocation QP structure (eigenvectors matrix %ld x %ld)\n",d,d);
	  return NULL;
	}
  }

  /* Eigenvalues vector */
  x->eval=(double *) malloc  (d*sizeof(double));
  if (!x->eval)
  {
	  QP_FreeQP(&x);
	  fprintf(stderr,"ERROR: allocation QP structure (eigenvalues vector %ld)\n",d);
	  return NULL;
  }
 
  return x;

}



/***********************************************/
/* Free QP structure					       */
/***********************************************/

void QP_FreeQP(QP_pQP *x)
{
  long i;

  if (x && (*x)) {
    if ((*x)->Q) {
	for (i=0;i<(*x)->d;i++)
	{
		if ((*x)->Q[i])
		{
			free((*x)->Q[i]);
			(*x)->Q[i] = NULL;
		}
	}
      free((*x)->Q);
      (*x)->Q=NULL;
    }
    if ((*x)->F) {
      free((*x)->F);
      (*x)->F=NULL;
    }
    if ((*x)->h) {
      free((*x)->h);
      (*x)->h=NULL; 
    }
   if ((*x)->evec) {
	for (i=0;i<(*x)->d;i++)
	{
		if ((*x)->evec[i])
		{
			free((*x)->evec[i]);
			(*x)->evec[i]=NULL;
		}
	}
	free((*x)->evec);
	(*x)->evec=NULL;
    }
    if ((*x)->eval) {
      free((*x)->eval);
      (*x)->eval=NULL;
    }
    free(*x);
    (*x)=NULL;
  }

}

/********************************************************
*			Means and variances computation				*
*********************************************************/
void QP_calcularMediasVarianzas(double **data, long nData,long d,double mv,double *mean, double *variance)
{
  
  long i,j,n;
  
  memset(&mean[0],'\0',sizeof(double)*d);
  memset(&variance[0],'\0',sizeof(double)*d);
 
  /* Mean */

  for (i=0;i<d;i++)
  {
	n=0;
	for (j=0;j<nData;j++)
	{
	  if (data[j][i]!=mv)
	  {
		n++;
		mean[i]+=data[j][i];
	  }
	}

	if (nData>0)
	  mean[i]/=n;	
	else
	  mean[i]=0.0;
  }


  /* Variance */



  for (i=0;i<d;i++)
  {
	n=0; 
	for (j=0;j<nData;j++)
	{  
  
	  if (data[j][i]!=mv)
	  {
		variance[i]+=(data[j][i]-mean[i])*(data[j][i]-mean[i]);
		n++;
	  }
	}
	variance[i]/=n;
  }


}

/********************************************************
*			Discretization of data						*
*********************************************************/
short QP_DataDiscretization(double **data, long n, long d, long nClasses, long nSegmentos,long nSigma)
{
  
  double lseg,ns,segmento=0;
  double *mean=NULL, *variance=NULL;
  long i,j;
  
  
  /* Memory allocation */
  mean=(double *)malloc(d*sizeof(double));

  if (!mean)
  {
	  fprintf(stderr,"ERROR: memory allocation of matrix 'mean' (%ld)\n",d);
	  return QP_ERROR;
  }

  variance=(double *)malloc(d*sizeof(double));
  if (!variance)
  {
	  fprintf(stderr,"ERROR: memory allocation of matrix 'variance' (%ld)\n",d);
	  free(mean);
	  return QP_ERROR;

  }
  
  
  QP_calcularMediasVarianzas(data,n,d,9999,mean,variance);
  

  /* asignacion de las longitudes de segmentos y ns */
  lseg=2.0*nSigma/(nSegmentos-2.0);


  /* en el caso de un valor constante el segmento es el cero */
  /* comprueba que no se esta ante un valor constante */
  for (i=0;i<n;i++)
  {
	// The first dimension represents the class.
	for (j=1;j<d;j++)
	{
	  if (variance[j]==QP_MAX_VARIANZA)
	  {
		data[i][j]=0;
	  }
	  else
	  {
		ns=(data[i][j]-mean[j])/sqrt(variance[j]);
		if (ns>nSigma) {
		  segmento=nSegmentos-1;
		} else if (ns<-nSigma) {
		  segmento=0;
		} else {
		  ns+=nSigma;
		  ns/=lseg;
		  segmento=(long)floor(ns+1);
		}
	  }
	  data[i][j]=segmento;
	}
  }
  
  
  free(mean);
  free(variance);
  return QP_OK;
 
}

/*********************************************************/
/* 			Similarity matrix							*/
/********************************************************/
#define TINY 1e-6

short QP_EntropyMatrix (QP_pQP qp,double **data, long nData, long d, long nClasses, long nSegment, long nSigma, long * aleat, long ns, double alpha)
{

  long	i,j,k,l;
  double **MI=NULL;
  long *count=NULL;
  long *total1=NULL,*total2=NULL;


	
  if (!QP_DataDiscretization(data,  nData, d, nClasses, nSegment, nSigma))
	return QP_ERROR;


  
  /* +1 sample: target */
  ns=ns+1;
  
  /* Allocation: mutual information matrix */
  MI = (double **) calloc (ns,sizeof(double*));

  if (!MI)
  {
	fprintf(stderr,"ERROR: memory allocation mutual information matrix (%ld x %ld)\n",ns,d);
	return QP_ERROR;

  }

  for  (i=0;i<ns;i++)
  {
	MI[i]=(double *)calloc(d,sizeof(double));

	if(!MI[i])
	{
		for (j=i-1;j>=0;j--)
		  free(MI[j]);
		free(MI);
		return QP_ERROR;
	}
  }

  /* MUTUAL INFORMATION FEATURE-CLASS */
  count = (long *) malloc (nClasses*nSegment*sizeof(long));

  if (!count)
  {
	fprintf(stderr,"ERROR: memory allocation matrix 'count' (%ld)\n",nClasses*nSegment);
	for (i=0;i<ns;i++)
	  free(MI[i]);
	free(MI);
	return QP_ERROR;

  }

  total1=(long *) malloc (nClasses*sizeof(long));
  if (!total1)
  {
	fprintf(stderr,"ERROR: memory allocation of matrix 'total1' (%ld)\n",nClasses);
	for (i=0;i<ns;i++)
	  free(MI[i]);
	free(MI);
	free(count);
	return QP_ERROR;

  }
 
  total2=(long *) malloc (nSegment*sizeof(long));
  if (!total2)
  {
	fprintf(stderr,"ERROR: memory allocation of matrix 'total2' (%ld)\n",nSegment);
	for (i=0;i<ns;i++)
	  free(MI[i]);
	free(MI);
	free(count);
	free(total1);
	return QP_ERROR;

  }

  for(j=1;j<d;j++)
  {
	for (k=0;k<(nClasses*nSegment);k++)
		count[k]=0.0;
	for (k=0;k<nData;k++)
	{
		count[(long)data[k][0]*nSegment+(long)data[k][aleat[j]]]++;
	}
	
	
	for(i=0;i<nClasses;i++) total1[i]=0;
	for(i=0;i<nSegment;i++) total2[i]=0;

	for(i=0;i<nClasses;i++)
	{
		for(k=0;k<nSegment;k++)
	
		{
			total1[i]+=count[i*nSegment+k];
			total2[k]+=count[i*nSegment+k];		
		}
	}

	/* Mutual Information approximation */
	/* Loop over combinations */
	for (i=0;i<nClasses;i++)
		for (k=0;k<nSegment;k++)
			if ((total1[i]>TINY) && (total2[k]>TINY) && (count[i*nSegment+k]>TINY))	
				MI[0][j]+=count[i*nSegment+k]*log(nData*(count[i*nSegment+k]/(1.0*total1[i]*total2[k])));
	MI[0][j]/=nData;

  }
   
  
  free(total1);
  free(total2);
  free(count);
  
   /* Linear term */
   for(i=1;i<d;i++)
   {	
	  qp->F[i-1]=alpha*fabs(MI[0][i]);
   }


  /* MUTUAL INFORMATION VARIABLE-VARIABLE */  
  count = (long *) malloc (nSegment*nSegment*sizeof(long));

  if (!count)
  {
	fprintf(stderr,"ERROR: memory allocation of matrix 'count' (%ld)\n",nSegment*nSegment);
	for (i=0;i<ns;i++)
	  free(MI[j]);
	free(MI);
	return QP_ERROR;

  }

  total1=(long *) malloc (nSegment*sizeof(long));
  if (!total1)
  {
	fprintf(stderr,"ERROR:memory allocation of matrix 'total1' (%ld)\n",nSegment);
	for (i=0;i<ns;i++)
	  free(MI[j]);
	free(MI);
	free(count);
	return QP_ERROR;

  }

  total2=(long *) malloc (nSegment*sizeof(long));
  if (!total2)
  {
	fprintf(stderr,"ERROR: memory allocation of matrix 'total2' (%ld)\n",nSegment);
	for (i=0;i<ns;i++)
	  free(MI[j]);
	free(MI);
	free(count);
	free(total1);
	return QP_ERROR;

  }


  for (i=1;i<ns;i++)
  {
	for(j=i;j<d;j++)
	{
		for (k=0;k<(nSegment*nSegment);k++)
			count[k]=0.0;
		for (k=0;k<nData;k++)
			count[(long)data[k][aleat[i]]*nSegment+(long)data[k][aleat[j]]]++;

		for(l=0;l<nSegment;l++) total1[l]=total2[l]=0;		
		
		for(l=0;l<nSegment;l++)
		{
			for(k=0;k<nSegment;k++)	
			{
				total1[l]+=count[l*nSegment+k];
				total2[k]+=count[l*nSegment+k];		
			}
		}
	

		/* Mutual Information approximation */
		/* Loop over combinations */
		for (l=0;l<nSegment;l++)
			for (k=0;k<nSegment;k++)
				if ((total1[l]>TINY) && (total2[k]>TINY) && (count[l*nSegment+k]>TINY))	
					MI[i][j]+=count[l*nSegment+k]*log(nData*(count[l*nSegment+k]/(1.0*total1[l]*total2[k])));
		MI[i][j]/=nData;

	}
  }

  /* Symmetric matrix */

  for (i=0;i<ns;i++)  
    for (j=0;j<i;j++)
      MI[i][j] = MI[j][i];


  /* OK */

  free(total1); 
  free(total2);
  free(count);
  
   /* Quadratic term */
  for (i=1;i<ns;i++)
  {
 	  for (j=1;j<d;j++)
 	  {	
 		  qp->Q[i-1][j-1]=(1.0-alpha)*fabs(MI[i][j]);			
 	  }
  }
  
  
  for (i=0;i<ns;i++)
	free(MI[i]);
  free(MI);
  
  return QP_OK;
  

}

#undef TINY

short QP_CorrelationMatrix (QP_pQP qp,double **data, long nData, long d, long nClasses, long * aleat, long ns, double alpha)
{
  
  double *mean=NULL, *variance=NULL,**C=NULL;
  long i,j,k;
  
  
  /* +1 sample: target */
  ns=ns+1;
  
  /* Memory allocation */
  mean=(double *)malloc(d*sizeof(double));

  if (!mean)
  {
	  fprintf(stderr,"ERROR: memory allocation of matrix 'mean' (%ld)\n",d);
	  return QP_ERROR;
  }

  variance=(double *)malloc(d*sizeof(double));
  if (!variance)
  {
	  fprintf(stderr,"ERROR: memory allocation of matrix 'variance' (%ld)\n",d);
	  free(mean);
	  return QP_ERROR;

  }
  
  
  QP_calcularMediasVarianzas(data,nData,d,9999,mean,variance);
  
  
  C = (double **) malloc (ns*sizeof(double*));

  if (!C)
  {
	  fprintf(stderr,"ERROR: memory allocation of matrix 'C'\n");
	  free(mean);
	  free(variance);
	  return QP_ERROR;
  }

  for (i=0;i<ns;i++)
  {
	  C[i]=(double *) malloc (d*sizeof(double));

	  if (!C[i])
	  {
		  fprintf(stderr,"ERROR: memory allocation of matrix 'C'\n");
		  free(mean);
		  free(variance);
		  for(j=0;j<i;j++)
			  free(C[j]);
		  free(C);
		  return QP_ERROR;

	  }
  }	

  
  
  // Computing correlation
  // VARIABLE-VARIABLE
  for (i=0;i<ns;i++)
  {    
	
	for (j=i;j<d;j++)
	{
	  C[i][j]=0.0;

	  for (k=0;k<nData;k++)
	  {				 
		  C[i][j]+=(data[k][aleat[i]]-mean[aleat[i]])*(data[k][aleat[j]]-mean[aleat[j]]);				
	  }
	  C[i][j]/=nData;
	}

	
  }
  

  
  for (i=0;i<ns;i++)
  {
	for (j=i;j<d;j++) {
	  if ((variance[aleat[i]]<=QP_VAR_TINY && variance[aleat[j]]>QP_VAR_TINY) || 
			  (variance[aleat[i]]>QP_VAR_TINY && variance[aleat[j]]<=QP_VAR_TINY)) {
			  C[i][j]=0.0;	   
	  } else if (variance[i]<=QP_VAR_TINY && variance[j]<=QP_VAR_TINY) {
		  /* Irene: dos constantes estan muy correlacionadas */
		  C[i][j]=1.0;	
		} else {
		  C[i][j]/=sqrt(variance[aleat[i]]*variance[aleat[j]]);
		}
	}
  }

 
 // CLASS-VARIABLE
 QP_WeightedCorrelationRandom(C,mean,variance,data,nData,d,nClasses,aleat);
 
 
  /* asigna la simetrica */

  /* Symmetric matrix */

  for (i=0;i<ns;i++)  
    for (j=0;j<i;j++)
      C[i][j] = C[j][i];

   /* Linear term */
   for(i=1;i<d;i++)
   {	
	  qp->F[i-1]=alpha*fabs(C[0][i]);
   }

		  
		  
  /* Quadratic term */
 
  for (i=1;i<ns;i++)
  {
 	  for (j=1;j<d;j++)
 	  {	
 		  qp->Q[i-1][j-1]=(1.0-alpha)*fabs(C[i][j]);			
 	  }
  }

  /* salida correcta */  
  free(mean);
  free(variance);
  for (i=0;i<ns;i++)
	free(C[i]);
  free(C);
  return QP_OK;
  
  
  
}

/***************************************
   *  Calculo de weigthed Pearson correlation coefficient */
  /* Se suponen las clases numeradas de 0 a nClases-1 */
  /* Paper: "Correlation-based feature selection for discrete and numeric
   * class machine learning"					*/
  /*********************************************************/
void QP_WeightedCorrelationRandom(double **C,double *media, double *varianza,double **data,long nData,long d,long nClases,long *aleat)
{
	long i,j,k;
	double correlacion;
	long *nk=NULL;

	/* vector de numero de puntos por clase*/
  	nk=(long *) calloc (nClases,sizeof(long));
  	if (!nk)
  	{
		fprintf(stderr,"ERROR QP_WeightedCorrelationRandom (vector nk)\n");
		return;
  	}

	for(i=0;i<nData;i++)
		nk[(long)data[i][0]]++;
		

	for (j=1;j<d;j++)
	{
		C[0][j]=0.0;
		for (i=0;i<nClases;i++)
		{
			media[0]=(double)(nk[i]/(1.0*nData));
			varianza[0]=(double)(nk[i]/(1.0*nData))-(media[0]*media[0]);	
			correlacion=0.0;			
			for (k=0;k<nData;k++)
			{

					if (data[k][0]==i)
						correlacion+=(1.0-media[0])*(data[k][aleat[j]]-media[aleat[j]]);
					else
						correlacion+=(0.0-media[0])*(data[k][aleat[j]]-media[aleat[j]]);
				
      		}
			correlacion/=nData;				

    		
			if ((varianza[0]<=QP_VAR_TINY && varianza[j]>QP_VAR_TINY) || 
      				(varianza[0]>QP_VAR_TINY && varianza[j]<=QP_VAR_TINY)) {	
       				correlacion=0.0;
      			} else if (varianza[0]<=QP_VAR_TINY && varianza[j]<=QP_VAR_TINY) {
					/* Irene: Constantes totalmente correlacionadas */
					correlacion=1.0;			
      			} else {				
       				correlacion/=sqrt(varianza[0])*sqrt(varianza[j]);				
      			}					
			
			C[0][j]+=fabs((nk[i]/(1.0*nData))*correlacion);			
    		}
	}

	free(nk);
  
}



/********************************************************************
*							QP SOLVER								*
*********************************************************************/
/***
* ns: number of subsamples in the Nystrom Method. -1 if Nystrom does not apply
***/

short QP_SolveQP (QP_pQP qp, long ns)
{

	double *C=NULL, *F=NULL, *h=NULL;
	long i,j;
	float cost=0;
	long rank;
	NYSTROM_pNYSTROM nys;

	if (!qp)
		return QP_ERROR;


	/* Eigenvalue descomposition */
	
	if (ns==-1)
	{
		if (QP_CalcEigen(qp)==QP_ERROR) 
		{
			return QP_ERROR;
		}
	}

	else
	{
		nys= NYSTROM_CalcularNYSTROM(qp->Q, qp->d, ns);

		if(!nys)
		{
			fprintf(stderr,"ERROR in Nystrom\n");
			return QP_ERROR;
		}

		/* copying the results in the structure */

		for (i=0;i<qp->d;i++) qp->eval[i]=0.0;
		for (i=0;i<ns;i++)
		{
			qp->eval[i]=nys->eval[i];		
		}

		for (i=0;i<qp->d;i++)
			for (j=0;j<ns;j++)
				qp->evec[i][j]=nys->evec[i][j];

		NYSTROM_LiberarNYSTROM(&nys); nys=NULL;

	}

	/* Matrix Rank Approximation */
	rank=qp->d;
	for (i=0;i<qp->d;i++)
	{
		// Matlab Rank Tol
		if (qp->eval[i]<(qp->eval[0]*EPSILON))
		{
			rank=i;
			break;
		}
	}


	/* Matrix formating */


	C = (double *) malloc (rank*rank*sizeof(double));		
	if (!C)
	{
		fprintf(stderr,"ERROR: memory allocation of matrix C (%ld)\n",rank*rank);
		return QP_ERROR;	
	}
	
	F = (double *) malloc (rank*sizeof(double));
	if (!F)
	{
		fprintf(stderr,"ERROR: memory allocation of vector F (%ld)\n",rank);
		free(C);
		return QP_ERROR;
	}



	/**** SUBSPACE PROBLEM ****/


	for (i=0;i<rank*rank;i++) C[i]=0.0;

	
	for (i=0;i<rank;i++)
	{		
		C[i*rank+i]=qp->eval[i];				
	}


	for (i=0;i<rank;i++) F[i]=0.0;

	for (i=0;i<rank;i++)
	{
		for (j=0;j<qp->d;j++)
		{
		 	F[i]+=qp->F[j]*qp->evec[j][i];
		}
	}


	/* Result is allocated in the function */			

	//printf("rango=%ld\n",rango);fflush(stdout);
	
	if (split_cake2(qp->d,rank,C,F,&h,&cost,-EPSILON,1.0+EPSILON,qp->evec)!=0)
	  return QP_ERROR;

	
	/* Result is copied in the QP struct coming back to the original space*/	
	for (i=0;i<qp->d;i++)
	{
		qp->h[i]=0.0;	
		for (j=0;j<rank;j++)
			qp->h[i]+=h[j]*qp->evec[i][j];


	}

	/* OK */
	free(C);
	free(F);
	free(h);
	return QP_OK;
}



/************************************************************
*				Eigenvalue descomposition					*
*************************************************************/

/**********************************************************************/
/**	      diagonalizacion de la matriz de covarianza          *********/
/**    salida 0: error de alocacion			                      *********/
/**	      1: salida correcta			                            *********/
/**	      2: no converge				                              *********/
/**********************************************************************/

short QP_CalcEigen(QP_pQP x)
{

  long		i,j;
  double *A=NULL,*V=NULL,*S=NULL;


  
  /* Se forma la matriz A (se supone de rango maximo) */   
  A = (double *) malloc ((x->d)*(x->d)*sizeof(double));


  if (!A)
  {
	fprintf(stderr,"ERROR: memory allocation of matrix A (%ld)\n",(x->d)*(x->d));
	return QP_ERROR;
  }


  /* Paso al formato contiguo */
  for (i=0;i<(x->d);i++)
  {
	for (j=0;j<(x->d);j++)
	{
		A[i*(x->d)+j]=x->Q[i][j];
	}
  }


	
  V=(double *) malloc (x->d*x->d*sizeof(double));

  if (!V)
  {
	fprintf(stderr,"ERROR: memory allocation of matrix V (%ld)\n",x->d*x->d);
	return QP_ERROR;
  }


  S=(double *) malloc(x->d*sizeof(double));
  
  if (!S)
  {
	fprintf(stderr,"ERROR: memory allocation of vector S (%ld)\n",x->d);
	return QP_ERROR;
  }


  // r: contiene el rango de la matriz 
  svd(x->d,A,V,S);
  

 
  
  /* Se mueve el resultado a la estructura */

  for(i=0;i<(x->d);i++)
	x->eval[i]=S[i];

  for (i=0;i<(x->d);i++)
  {

	for (j=0;j<(x->d);j++)		
		x->evec[i][j]=V[j*(x->d)+i];

  }
  
  free(A);
  free(V);
  free(S);

  return QP_OK;
  
}

/*** 
* LOOK: Evec=NULL if basis matrix is not required
* m: original problem dimension
* n: dimension in the subspace
***/

int split_cake2(int d,int n,double *C,double *F,double * *res,float *fr, double min_alo, double max_alo, double **evec) {


  int m, p, nact,ierr;
  int i,j;
  int n2=n*n;
  double crval;  
  //double *G, g0[n], *CI, ci0[d+1], x[n], iact[d+1];
  double *G, *g0, *CI, *ci0, *x, *iact;

  int iter[2];
  double* work;
  long dwork;



  //dwork=(d2+11*d);

  dwork=n2+5*n+2*d+1;

  work = (double *) malloc (dwork*sizeof(double));



  if (!work)
  {
	puts("No enough memory");
	exit(0);

  }

  (*res)=(double *)malloc(sizeof(double)*n);

  if (*res==NULL) {
    puts("No enough memory");
    exit(0);
  }

  G=(double *)malloc(sizeof(double)*n2);
  if (G==NULL) {
    puts("No memory for G in <split_cake2>");
    exit(0);
  }
  
  // IRENE: n restricciones de minimo y una de igualdad
  //CI=(double *)malloc(sizeof(double)*(n2+n));  
  // OJO
  CI=(double *)malloc(sizeof(double)*((d+1)*n)); 
  if (CI==NULL) {
    puts("No memory for CI in <split_cake2>");
    exit(0);
  }

  g0=(double*) malloc (n*sizeof(double));
  if (g0==NULL)
  {
    puts("No memory for g0 in <split_cake2>");
    exit(0);
  }

  ci0=(double*) malloc((d+1)*sizeof(double));
  if (ci0==NULL)
  {
    puts("No memory for ci0 in <split_cake2>");
    exit(0);
  }

  x=(double *) malloc (n*sizeof(double));
  if (x==NULL)
  {
    puts("No memory for x in <split_cake2>");
    exit(0);
  }

  iact=(double *)malloc((d+1)*sizeof(double));
  if (iact==NULL)
  {
    puts("No memory for iact in <split_cake2>");
    exit(0);
  }






  for(i=0;i<n*n;++i) 
      G[i]=C[i];

  for(i=0;i<n;++i) 
    g0[i]=F[i];



  /* No se imponen condiciones de igualdad */  
  //m = 0;
  m=1;

  //for(i=0;i<(n2+n);++i) CI[i]=0.0;
  for(i=0;i<((d+1)*n);++i) CI[i]=0.0;

    /* Problem resolved in the original space */
  if (evec==NULL)
  {

	/* Condicion de que todas las variables sumen 1 */
	for (i=0;i<n;i++) CI[i]=1.0;
	ci0[0]=1.0;

	for(i=0;i<2*n2;++i) CI[i]=0.0;
	for(i=0;i<n;++i) {
		CI[i+n*i]=1.0;
		ci0[i]=min_alo;
	}
/*	
	for(i=n;i<2*n;++i) {
		CI[(i-n)+n*i]=-1.0;
		//ci0[i]=max_alo;
		// IRENE: modificado para qpgen2
		ci0[i]=-max_alo;
	}
*/
  }
  else  /* Problema equivalente diagonalizado */
  {

	/* Condicion de que todas las variables sumen 1 */
	for (i=0;i<n;i++)
		//for (j=0;j<n;j++)
		//OJO
		for (j=0;j<d;j++)
			CI[i]+=evec[j][i];
	ci0[0]=1.0;

	/* Transformacion de las condiciones en el problema diagonal */

		
	//for(i=1;i<=n;++i) {
	for(i=1;i<=d;++i) {		
		for (j=1;j<=n;j++)
		{
			//CI[(j-1)+i*n]=evec[i-1][j-1];
			CI[(j-1)+i*n]=evec[i-1][j-1];
			//printf("%ld\n",(j-1)+i*n);
		}
		ci0[i]=min_alo;
	}
  }

  //p=n+1;
  p=d+1;
  

  //*fr= solve_quadprog(G, g0, n, CE, ce0, m, CI, ci0, p, x) ;
  ierr=0;

/*
  qpgen2(dmat, dvec, fddmat, n, sol, crval, amat, 
     *     bvec, fdamat, q, meq, iact, nact, iter, work, ierr)  
*/


  qpgen2_(G, g0, &n, &n,x,&crval,CI,ci0,&n,&p,&m,iact, &nact,iter, work,&ierr);


  if (ierr!=0)
	fprintf(stdout,">>> ERROR solving the QP problem\n");
  else
	fprintf(stdout,">>> OK solving the QP problem\n");
  fflush(stdout);



/*** DEBUG ***/
/*
  printf("crval=%lf\n",crval);
  printf("ierr=%lf\n",ierr);
  printf("iter[0]=%d iter[1]=%d\n",iter[0],iter[1]);
  printf("nact=%d\n",nact);


  printf("*** Solucion con restricciones ***\n");
  for (i=0;i<n;i++)
	printf("%lf ",x[i]);
  printf("\n");


  printf("*** Solucion sin restricciones ***\n");
  for (i=0;i<n;i++)
	printf("%lf ",g0[i]);
  printf("\n");

  printf("*** iact ***\n");
  for (i=0;i<nact;i++)
	printf("%d ",iact[i]);
  printf("\n");

*/
/*** DEBUG ***/

 for(i= 0; i < n; i++) 
    (*res)[i]=x[i];

  free(work);
  free(CI);
  free(G);
  free(g0);
  free(ci0);
  free(x);
  free(iact);
 
 return ierr;

}



/********************************************************************
*				WRITE THE QPFS SOLUTION 							*
*********************************************************************/
short QP_WriteQP(QP_pQP qp, char *FicheroSal, long submuestreos, long *aleat, long *sort, long offset)
{

  FILE *fSal=NULL;
  long i;
  

  if (!(fSal=fopen(FicheroSal,"w"))) {
    fprintf(stderr,"ERROR opening the file '%s' for writting\n",FicheroSal);
    return QP_ERROR;
  }

  if (qp->alpha==0.0)
  {
  	for (i=0;i<qp->d;i++)
		fprintf(fSal,"%ld\t%ld\t%10.17le\t\t%10.17le\n",i,(aleat[sort[i]+1]-1)+offset,qp->h[i],0.0);
  }
  else
  {
  	for (i=0;i<qp->d;i++)
		fprintf(fSal,"%ld\t%ld\t%10.17le\t\t%10.17le\n",i,(aleat[sort[i]+1]-1)+offset,qp->h[i],qp->F[i]/qp->alpha);
  }


  /* salida correcta */
  fclose(fSal);
  return QP_OK;

}

/************************************************************
*					SORT THE QPFS SOLUTION					*
*************************************************************/

/************************************************************
*			Ordenacion de variables				*
* //FIXME Ordenacion por burbuja					*
* Se supone que la permutacion tiene el orden de elementos 	*
*************************************************************/

void QP_SortQP(QP_pQP qp, long *permutacion)
{

	long i;
	long aux;
	double aux_h;
	short swapped;
	long n_positivos;
	
	if (!permutacion)
		return;


	/* Ordenacion por la solucion de QPFS */
	do
	{
		swapped=0;
		for  (i=0;i<(qp->d-1);i++)
		{
			/* Irene: en caso de empate se ordenan por similitud con la clase */
			if (qp->h[i] < qp->h[i+1])
			{

				aux_h=qp->h[i];
				qp->h[i]=qp->h[i+1];
				qp->h[i+1]=aux_h;

				aux_h=qp->F[i];
				qp->F[i]=qp->F[i+1];
				qp->F[i+1]=aux_h;


				aux=permutacion[i];
				permutacion[i]=permutacion[i+1];
				permutacion[i+1]=aux;
			
				swapped=1;
			}

		}
	}while(swapped);

	n_positivos=0;

	while((n_positivos<qp->d)&&(fabs(qp->h[n_positivos])>EPSILON))
	{
		n_positivos++;
	}

	/* Ordenacion por F's en caso de h=0 */

	do
	{
		swapped=0;
		for  (i=n_positivos;i<(qp->d-1);i++)
		{

			/* Irene: en caso de empate se ordenan por similitud con la clase */

			if (qp->F[i] < qp->F[i+1])
			{

				aux_h=qp->h[i];
				qp->h[i]=qp->h[i+1];
				qp->h[i+1]=aux_h;

				aux_h=qp->F[i];
				qp->F[i]=qp->F[i+1];
				qp->F[i+1]=aux_h;


				aux=permutacion[i];
				permutacion[i]=permutacion[i+1];
				permutacion[i+1]=aux;
			
				swapped=1;
			}

		}
	}while(swapped);
	
	
		
}


