/**************************************************************/
/************* QPFS.c ****************************************/
/**************************************************************/
/*    Irene Rodriguez-Lujan    *********************************/
/*    Ramon Huerta			  *********************************/
/*    Charles Elkan			  *********************************/
/*    Carlos Santa Cruz		  *********************************/
/**************************************************************/
/** June 3rd, 2010                                        ****/
/** Based on "Quadratic Programming Feature Selection"       **/
/**************************************************************/

    
#include <stdio.h>   
#include <stdlib.h>       
#include <math.h>       
#include <time.h>
#include <string.h>
#include <unistd.h>
#include <limits.h>
#include "QP.h"
 

long * QPFS_getFileInfo(FILE *fInput,long *nData,long *d,long *nClasses,long *offset);

#define LINE_LENGTH			100000
    
  
#define HELP "\n\n\
  Copyright 2010 Ramon Huerta, Charles Elkan, Carlos Santa Cruz and Irene Rodriguez-Lujan [irene.rodriguez@uam.es]\n\n\
  Free software for non-commercial research use; for all other uses please contact the authors\n\n\
  Usage: %s -F <dataset> -O <output file> [optional arguments]\n\
  -h  Help. Print this message.\n\
  -F  <dataset>\t\t  File containing n rows and d+1 columns -separated by space-, row - sample, column - variable/attribute. First colum represents the target value.\n\
  -O  <output file>\t\t File containing the rank of the features according to the QPFS criteria (default standard output).\n\
  -p  <rate of subsampling in Nystrom method>\t\t A float number in [0,1] or {-1} if Nystrom method is not applied.(default -1)\n\
  -a  <alpha parameter>\t\t Value of the parameter alpha in the QPFS method (default 0.5)\n\
  -C  <correlation flag>\t\t To use correlation as similarity measure (default mutual information)\n\
  -s  <number of standard deviation>\t\t Number of standard deviation of data to cover by the discretization process (default 1)\n\
  -n  <number of segments>\t\t Number of segments in the discretization process [must be an odd number] (default 3)\n\
  -E  <alpha estimation>\t\t Flag to estimate the alpha parameter (default NO)\n\n\
\n\n\n\
*** This program was developed for the paper \"Quadratic Programming Feature Selection\",\n\
	 Irene Rodriguez-Lujan, Ramon Huerta, Charles Elkan, and Carlos Santa Cruz,\n\
     Journal of Machine Learning Research,\n\
	 Vol. 11(Apr): 1491−1516, 2010.\n\n"

 
#define FREEALL		{if (data) for (i=0;i<nData;i++) free(data[i]);free(data);\
		   if (fOutput) fclose(fOutput);\
		   if (qp) QP_FreeQP(&qp);\
		   if (aleat) free(aleat);\
		   if (permutacion) free(permutacion);\
		   if (map) free(map);}


/*
 * External variables
 */  


extern int optind;
extern int optopt; 
extern int opterr;
extern char *optarg;
    

/**********************************************************************
 *      main()
 */ 
   
int main (int argc, char **argv)   

{        
   
  FILE *fInput=NULL, *fOutput=NULL;
  char inputFile[500],outputFile[500];
  long nData,d,i,j;
  long aux;
  long *map=NULL;
  long offset=0;
  int c;
  QP_pQP qp=NULL;
  double **data=NULL;
  double subsampling_rate;
  long ns;
  double alpha;
  long *aleat=NULL,*permutacion=NULL;
  short cor;
  long nClasses,nSegments,nSigmas;
  long nvar;
  double valor;
  char auxiliar[LINE_LENGTH], *pw=NULL;
  short alpha_estimation;
  double mean_q, mean_f;
  short flag;
  
  /*
   * Default values
   */

  outputFile[0]='\0';
  inputFile[0]='\0';
  subsampling_rate=-1.0;			// default: no Nystrom
  alpha=0.5;
  cor=0;
  nClasses=2;
  nSegments=3;
  nSigmas=1;
  alpha_estimation=0;
  
 
  /* Cheking the number of parameters */
  if (argc<4) {
	fprintf(stderr,HELP,argv[0]);
    return 1;
  }

  
  /* command line options */
  
  optind=1;
  while ((c=getopt(argc,argv,"hF:O:p:a:c:Cs:n:E"))!=EOF) {
    switch (c) {
    case 'h':
      fprintf(stderr,HELP,argv[0]);
      return 1;
    case '?':
      fprintf(stderr,HELP,argv[0]);
      printf("Unrecognized parameter. Use '%s -h'\n",argv[0]);
      return 1;
    case 'F':
      strcpy(inputFile,optarg);
      break;
	case 'O':
      strcpy(outputFile,optarg);
      break;
    case 'p': 
      subsampling_rate=atof(optarg);
      break;  
    case 'a': 
      alpha=atof(optarg);
      break;
    case 'c': 
      nClasses=atol(optarg);
      break;
    case 'C': 
      cor=1;
      break;
	case 's': 
      nSigmas=atol(optarg);
      break;
	case 'n': 
      nSegments=atol(optarg);
      break;
	case 'E': 
      alpha_estimation=1;
      break;
    default:
      fprintf(stderr,HELP,argv[0]);
      return 1;
    }
  }


  /* Parameter correctness*/
  if ((inputFile[0]=='\0') || (outputFile[0]=='\0') || (alpha<0.0) || (alpha>1.0) || 
	  ((subsampling_rate!=-1)&&((subsampling_rate<=0.0) || (subsampling_rate>=1.0)))
	  || (nSigmas<=0) || (nSegments<=0) || (nSegments%2!=1)) 
  {
    printf(HELP,argv[0]);
    return 1;
  }

  /* Read dataset file */
  if (!(fInput=fopen(inputFile,"r"))) {
    FREEALL;
	fprintf(stderr,"ERROR: cannot open '%s' file\n",inputFile);
    return 1;
  }

 
  map=QPFS_getFileInfo(fInput,&nData,&d,&nClasses,&offset);

  if (!map)
  {
	FREEALL; 
	return 1;
  }
  
  /* the first variable is the target class */
  offset--;
  /* print the dataset information */
  printf(">>> Dataset information:\n");
  printf("* Number of patterns: %ld\n",nData);
  printf("* Dimension: %ld\n",d);
  printf("* Number of classes: %ld\n\n\n",nClasses);
  fflush(stdout);
  
  
  
   /* Memory allocation (+1 target) */
  data=(double **) malloc (nData*sizeof(double *));
  if (!data)
  {
	FREEALL;
	fprintf(stderr,"ERROR: allocation of data matrix (%ld x %ld)\n",nData,(d+1));
	return 1;
  }
  
  for (i=0;i<nData;i++)
  {
	data[i]=(double *) calloc ((d+1),sizeof(double));
	if (!data[i])
	{
	  FREEALL;
	  fprintf(stderr,"ERROR: allocation of data matrix (%ld x %ld)\n",nData,(d+1));
	  return 1;
	}
  }
  

  rewind(fInput);

  for (i=0;i<nData;i++)
  {

	/* Lectura del target */
	if (fscanf(fInput,"%lf",&(data[i][0]))!=1) {
        fprintf(stderr,"Error al leer el target del patron %ld\n",i);
        FREEALL;
        return 1;
	}
	
	/* seraching the target in the map */
	flag=0;
	for (j=0;(j<nClasses)&&(flag==0);j++)
	{
	  if ((long)data[i][0]==map[j]){
		flag=1;
		data[i][0]=j;
	  }
	}
	
		

	if (!fgets(auxiliar,LINE_LENGTH,fInput))
	{
		printf("ERROR: reading pattern %ld\n",i);
		FREEALL;
		return 1;
	}

	pw=auxiliar;
	while((pw!=NULL) && (sscanf(pw,"%ld:%lf",&nvar,&valor)==2))
	{
		

		data[i][nvar-offset]=valor;

		/* encuentra el primer digito */
		while (*pw!='\0' && *pw==' ') pw++; 
		pw=strstr(pw," ");			
	}
	  
	
  }


  fclose(fInput);
  
  
  /* QP allocation */
  qp=QP_AllocQP(d,alpha);

  if (!qp)
  {
	FREEALL;
	fprintf(stderr,"ERROR: allocation QP structure\n");
	return 1;
  }


  /* random permutation of features for the Nystrom method */
  /* (d+1) to include the target */
  aleat=(long *) malloc ((d+1)*sizeof(long));

  if (!aleat)
  {
	  FREEALL;
	  fprintf(stderr,"ERROR al reservar el array 'aleat'\n");
	  return 1;
  }
  
  permutacion=(long *) malloc (d*sizeof(long));

  if (!permutacion)
  {
	  FREEALL;
	  fprintf(stderr,"ERROR al reservar el array 'permutacion'\n");
	  return 1;
  }
  
  /* initialization */
  for (i=0;i<(d+1);i++)
	aleat[i]=i;
  
  for (i=0;i<d;i++)
	permutacion[i]=i;

  /* initialization number subsamples */
  if  (subsampling_rate==-1)
  {
	ns=d;
  }
  else
  {
	  ns=(long)(ceil(subsampling_rate*d));
	  
	  fprintf(stdout, ">>> Number of subsamples in Nystrom method: %ld / %ld\n",ns,d);fflush(stdout);
	/* random permutation (the first element is fixed because it is associated with the target */
	for (i=1;i<(d-1);i++)
	{
		j=1+i+rand()%(d-1-i);
		aux=aleat[i];
		aleat[i]=aleat[j];
		aleat[j]=aux;
	}
  }

  if (alpha_estimation==1)
	alpha=0.5;

  /****************** SIMILARITY MATRIX ************************/
  if (!cor)
  {
	  printf(">>> Similarity measure: mutual information\n");fflush(stdout);
	  /* mutual information */
	  if (QP_EntropyMatrix(qp,data, nData, d+1, nClasses, nSegments, nSigmas, aleat,ns,alpha)==QP_ERROR)
	  {
		fprintf(stderr,"ERROR: call to QP_EntropyMatrix\n");
		FREEALL;
		return 1;
	  }
  }
  
  else
  {
	  printf(">>> Similarity measure: correlation\n");fflush(stdout);
	  /* Correlation */
	  if  (QP_CorrelationMatrix(qp,data,nData,d+1,nClasses, aleat, ns,alpha)==QP_ERROR)
	  {
		fprintf(stderr,"ERROR: call to QP_CorrelationMatrix\n");
		FREEALL;
		return 1;
	  }
  }


  /****************** QUADRATIC PROGRAMMING *************************/
  if (alpha_estimation==0)
  {
	if (alpha<1.0)
	{
	  if (subsampling_rate==-1)
		i=QP_SolveQP(qp,-1);
	  else
		i=QP_SolveQP(qp,ns);
	}
	else
	{
	  /* alpha=1.0 --> MaxRel */
	  for(i=0;i<qp->d;i++)
		  qp->h[i]=0.0;
	}
	 /* Variable ranking according to the QPFS coefficient */
	QP_SortQP(qp, permutacion);
  
	/* Escritura del problema en el fichero de salida */
	QP_WriteQP(qp, outputFile, ns,aleat,permutacion,offset+1);
  }
  else
  {
	mean_q=mean_f=0;
	for (i=0;i<ns;i++)
	{
	  for (j=0;j<d;j++)
		mean_q+=qp->Q[i][j];
	}
	mean_q=(mean_q)/(ns*d);
	for (i=0;i<d;i++)
	  mean_f+=qp->F[i];
	mean_f=(mean_f)/d;

	fprintf(stdout,">>> Alpha estimation=%lf\n",mean_q/(mean_q+mean_f));fflush(stdout);
	
  }




  /* salida del programa */
  FREEALL;
  return 0;
  
}  



/****************************************************
*		Obtaining information 						*
*****************************************************/

long * QPFS_getFileInfo(FILE *fInput,long *nData,long *d,long *nClasses,long *offset)
{
 
  long i;
  long *map=NULL,nvar;
  double valor;
  double target;
  short flag;
   char auxiliar[LINE_LENGTH], *pw=NULL;
  
  rewind(fInput);
  
  *nData=0;
  *nClasses=0;
  *d=LONG_MIN;
  *offset=LONG_MAX;
  
  
  
  while (fscanf(fInput,"%lf",&target)==1)
  {

	/* Searching the class in the map */
	flag=0;
	for (i=0;(i<(*nClasses))&&(flag==0);i++)
	{
	  if (map[i]==(long)target)
		flag=1;
	}
	
	if (flag==0)
	{
	  *nClasses=(*nClasses)+1;
	  map=(long *) realloc (map,(*nClasses)*sizeof(long));
	  if (!map)
	  {
		fprintf(stderr,"ERROR: reallocating target map\n");
		return NULL;
	  }
	  map[(*nClasses)-1]=(long)target;
	}

	if (!fgets(auxiliar,LINE_LENGTH,fInput))
	{
	  printf("ERROR: reading pattern %ld\n",i);
	  return NULL;
	}

	pw=auxiliar;
	while((pw!=NULL) && (sscanf(pw,"%ld:%lf",&nvar,&valor)==2))
	{	
	
	  if (nvar<(*offset))
		*offset=nvar;
	  
	  /* encuentra el primer digito */
	  while (*pw!='\0' && *pw==' ') pw++; 
	  pw=strstr(pw," ");			
	}
	if (nvar>(*d))
	  *d=nvar;
	*nData=(*nData)+1;  

  }

  (*d)=(*d)-(*offset)+1;

  
  return map;

  
}
